

structure Syntax :>
sig

    type name = string

    (* given a base name (e.g. 'x'), append a globally unique suffix 
       to make a 'fresh' name that is different from any names
       that have been generated before
       *)
    val freshName : string -> name
        
    datatype typ = Number | String;

    datatype exp = 
	Var of name
      | Num of int
      | Str of string
      | Plus of exp * exp
      | Cat of exp * exp
      | Let of exp * name * exp

    (* subst (e, (e',x)) implements capture-avoiding substitution e[e'/x] *)
    val subst : exp * (exp * name) -> exp
        
end
=
struct

    type name = string

    local 
	val suffix = ref 0 
    in
	fun freshName s = 
	     (suffix := (!suffix)+1;
	      s^(Int.toString(!suffix)))
    end
        
    datatype typ = Number | String;

    datatype exp = 
	Var of name
      | Num of int
      | Str of string
      | Plus of exp * exp
      | Cat of exp * exp
      | Let of exp * name * exp

    fun swap (e : exp, (x : name, y : name)) = raise Fail "unimplemented"

    fun subst (e, (e',x)) = raise Fail "unimplemented"
        
end